/**
 * @file yac_interface.c
 *
 * @copyright Copyright  (C)  2013 Moritz Hanke <hanke@dkrz.de>
 *                                 Rene Redler <rene.redler@mpimet.mpg.de>
 *                                 Teresa Holfeld <teresa.holfeld@zmaw.de>
 *
 * @author Moritz Hanke <hanke@dkrz.de>
 *         Rene Redler <rene.redler@mpimet.mpg.de>
 *         Teresa Holfeld <teresa.holfeld@zmaw.de>
 */
/*
 * Keywords:
 * Maintainer: Moritz Hanke <hanke@dkrz.de>
 *             Rene Redler <rene.redler@mpimet.mpg.de>
 *             Teresa Holfeld <teresa.holfeld@zmaw.de>
 * URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
 *
 * This file is part of YAC.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are  permitted provided that the following conditions are
 * met:
 *
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * Neither the name of the DKRZ GmbH nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "ensure_array_size.h"
#include "event.h"
#include "interpolation.h"
#include "yac_interface.h"
#include "utils.h"
#include "version.h"
#include "interp_grid.h"
#include "instance.h"
#include "fields.h"
#include "couple_config.h"
#include "config_xml.h"

int const YAC_LOCATION_CELL = CELL;
int const YAC_LOCATION_CORNER = CORNER;
int const YAC_LOCATION_EDGE = EDGE;

struct user_input_data_component {
  struct yac_instance * instance;
  size_t component_index;
};

struct user_input_data_points {
  int default_mask_id;
  struct yac_basic_grid * grid;
  enum yac_location location;
  size_t coordinates_idx;
};

struct user_input_data_masks {
  struct yac_basic_grid * grid;
  enum yac_location location;
  size_t masks_idx;
};

struct couple_config_transient {
  int global_comp_id;
  struct yac_couple_config * couple_config;
  size_t couple_idx;
  size_t transient_couple_idx;
};

static int default_instance_id = INT_MAX;

static int yac_instance_count = 0;

void ** pointer_lookup_table = NULL;
static int pointer_lookup_table_size = 0;

struct yac_basic_grid ** grids = NULL;
size_t num_grids = 0;

static struct user_input_data_component ** components = NULL;
static size_t num_components = 0;

static struct user_input_data_points ** points = NULL;
static size_t num_points = 0;

static struct user_input_data_masks ** masks = NULL;
static size_t num_masks = 0;

static struct couple_config_transient ** transients = NULL;
static size_t num_transients = 0;

/**
 * gives a unique index for a given pointer
 * @param[in] pointer
 * @return unique value associated to pointer
 * @see unique_id_to_pointer
 */
static int yac_pointer_to_unique_id(void * pointer) {

  pointer_lookup_table =
    xrealloc (pointer_lookup_table,
      (pointer_lookup_table_size + 1) * sizeof(*pointer_lookup_table));

  pointer_lookup_table[pointer_lookup_table_size] = pointer;

  pointer_lookup_table_size++;

  return pointer_lookup_table_size - 1;
}

/**
 * gives the pointer that is associated to the given id
 * @param[in] id      unique index previously returned by pointer_to_unique_id
 * @param[in] id_name name of the id
 * @return pointer the is associated to the given id \n NULL if the id is invalid
 */
static void * yac_unique_id_to_pointer(int id, char const * id_name) {

  YAC_ASSERT_F(
    id < pointer_lookup_table_size,
    "ERROR(yac_unique_id_to_pointer): invalid %s", id_name)
  return pointer_lookup_table[id];
}

/**
 * gives the unique index associated with pointer
 * @param[in] pointer
 * @return unique value associated to pointer
 * @remark the pointer has to have been registers using
 *         \ref yac_pointer_to_unique_id, otherwise this routine returns
 *         UINT_MAX
 */
static int yac_lookup_pointer(void const * pointer) {

  int ret = UINT_MAX;
  for (int i = 0; (ret == UINT_MAX) && (i < pointer_lookup_table_size); ++i)
    if (pointer_lookup_table[i] == pointer) ret = i;
  return ret;
}

/**
 * frees all memory used for the pointer/unique_id conversion
 * \remarks this should only be called after the last call to
 *          \ref yac_pointer_to_unique_id and \ref yac_unique_id_to_pointer, because afterwards
 *          \ref yac_unique_id_to_pointer will not be able to return the respective pointers
 *          for previously valid unique ids
 */
static void yac_free_pointer_unique_lookup() {

  free(pointer_lookup_table);
  pointer_lookup_table = NULL;
  pointer_lookup_table_size = 0;
}

/* ---------------------------------------------------------------------- */

static int yac_add_grid(
  const char * grid_name, struct basic_grid_data grid_data) {

  for (size_t i = 0; i < num_grids; ++i)
    YAC_ASSERT(
      strcmp(yac_basic_grid_get_name(grids[i]), grid_name),
      "ERROR(yac_add_grid): multiple definitions of grid with identical name")

  YAC_ASSERT(
    strlen(grid_name) <= YAC_MAX_CHARLEN,
    "ERROR(yac_add_grid): grid name is too long (maximum is YAC_MAX_CHARLEN)")

  struct yac_basic_grid * grid = yac_basic_grid_new(grid_name, grid_data);

  grids = xrealloc(grids, (num_grids + 1) * sizeof(*grids));
  grids[num_grids] = grid;
  num_grids++;
  return yac_pointer_to_unique_id(grid);
}

static void yac_free_grids() {

  for (size_t i = 0; i < num_grids; ++i)
    yac_basic_grid_delete(grids[i]);
  free(grids);
  grids = NULL;
  num_grids = 0;
}

/* ---------------------------------------------------------------------- */

static void yac_free_points() {

  for (size_t i = 0; i < num_points; ++i) free(points[i]);
  free(points);
  points = NULL;
  num_points = 0;
}

/* ---------------------------------------------------------------------- */

static void yac_free_masks() {

  for (size_t i = 0; i < num_masks; ++i) free(masks[i]);
  free(masks);
  masks = NULL;
  num_masks = 0;
}

/* ---------------------------------------------------------------------- */

static void yac_free_components() {

  for (size_t i = 0; i < num_components; ++i) free(components[i]);
  free(components);
  components = NULL;
  num_components = 0;
}

/* ---------------------------------------------------------------------- */

static void yac_free_transients() {

  for (size_t i = 0; i < num_transients; ++i) free(transients[i]);
  free(transients);
  transients = NULL;
  num_transients = 0;
}

/* ---------------------------------------------------------------------- */

static void check_default_instance_id(char const * routine_name) {

  YAC_ASSERT_F(
    default_instance_id != INT_MAX,
    "ERROR(%s): no default yac instance is defined yet", routine_name)
}

/* ---------------------------------------------------------------------- */

void yac_cinit_comm_group_instance (
  const char * xml_filename, const char * schema_filename, MPI_Comm comm,
  const char * group_name, int * yac_instance_id) {

  yac_mpi_init(comm);

  *yac_instance_id =
    yac_pointer_to_unique_id(
      yac_instance_new(
        yac_couple_config_xml_new(xml_filename, schema_filename),
        comm, group_name));
  yac_instance_count++;
}

void yac_cinit_comm_instance (
  const char * xml_filename, const char * schema_filename, MPI_Comm comm,
  int * yac_instance_id) {

  yac_cinit_comm_group_instance (
    xml_filename, schema_filename, comm, NULL, yac_instance_id);
}

void yac_cinit_comm_group_instance_f2c (
  const char * xml_filename, const char * schema_filename, MPI_Fint comm,
  char const * group_name, int * yac_instance_id) {

  yac_cinit_comm_group_instance(
    xml_filename, schema_filename, MPI_Comm_f2c(comm), group_name,
    yac_instance_id);
}

void yac_cinit_comm_instance_f2c (
  const char * xml_filename, const char * schema_filename, MPI_Fint comm,
  int * yac_instance_id) {

  yac_cinit_comm_group_instance(
    xml_filename, schema_filename, MPI_Comm_f2c(comm), NULL, yac_instance_id);
}

void yac_cinit_comm_group ( const char * xml_filename,
                            const char * schema_filename,
                            MPI_Comm comm,
                            const char * group_name ) {

  YAC_ASSERT(
    default_instance_id == INT_MAX,
    "ERROR(yac_cinit_comm_group): default yac instance already defined")

  yac_cinit_comm_group_instance(
    xml_filename, schema_filename, comm, group_name, &default_instance_id);
}

void yac_cinit_comm ( const char * xml_filename, const char * schema_filename,
                      MPI_Comm comm ) {

  YAC_ASSERT(
    default_instance_id == INT_MAX,
    "ERROR(yac_cinit_comm): default yac instance already defined")

  yac_cinit_comm_group_instance(
    xml_filename, schema_filename, comm, NULL, &default_instance_id);
}

void yac_cinit_comm_f2c ( const char * xml_filename, const char * schema_filename,
                          MPI_Fint comm_f ) {

  yac_cinit_comm(xml_filename, schema_filename, MPI_Comm_f2c(comm_f));

  return;
}

void yac_cinit_comm_group_f2c ( const char * xml_filename,
                                const char * schema_filename,
                                MPI_Fint comm_f,
                                const char * group_name ) {

  yac_cinit_comm_group(
    xml_filename, schema_filename, MPI_Comm_f2c(comm_f), group_name);

  return;
}

void yac_cinit_group_instance(
  const char * xml_filename, const char * schema_filename,
  const char * group_name, int * yac_instance_id) {

  yac_cinit_comm_group_instance(
    xml_filename, schema_filename, MPI_COMM_WORLD, group_name, yac_instance_id);
}

void yac_cinit_instance(
  const char * xml_filename, const char * schema_filename,
  int * yac_instance_id) {

  yac_cinit_comm_group_instance(
    xml_filename, schema_filename, MPI_COMM_WORLD, NULL, yac_instance_id);
}

void yac_cinit_group ( const char * xml_filename,
                       const char * schema_filename,
                       const char * group_name ) {

  yac_cinit_comm_group(
    xml_filename, schema_filename, MPI_COMM_WORLD, group_name);
}

void yac_cinit ( const char * xml_filename, const char * schema_filename ) {

  yac_cinit_comm_group(xml_filename, schema_filename, MPI_COMM_WORLD, NULL);
}

void yac_cinit_comm_dummy ( MPI_Comm comm ) {

  yac_mpi_init(comm);

  yac_instance_dummy_new(comm);
}

void yac_cinit_comm_dummy_f2c ( MPI_Fint comm_f ) {

  yac_cinit_comm_dummy(MPI_Comm_f2c(comm_f));
}

void yac_cinit_dummy() {

  yac_cinit_comm_dummy(MPI_COMM_WORLD);
}

/* ---------------------------------------------------------------------- */

static void yac_ccleanup_instance_(int yac_instance_id) {

  /* free memory */
  yac_instance_delete(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"));
  --yac_instance_count;

}

void yac_ccleanup_instance (int yac_instance_id) {

  yac_ccleanup_instance_(yac_instance_id);

  /* cleanup mpi */
  yac_mpi_cleanup();
}

void yac_ccleanup () {

  check_default_instance_id("yac_ccleanup");

  yac_ccleanup_instance(default_instance_id);
  default_instance_id = INT_MAX;
}

/* ---------------------------------------------------------------------- */

static void cleanup() {
  yac_mpi_finalize();
  yac_free_pointer_unique_lookup();
  yac_free_grids();
  yac_free_points();
  yac_free_masks();
  yac_free_components();
  yac_free_transients();
}

void yac_cfinalize_instance(int yac_instance_id) {

  yac_ccleanup_instance_(yac_instance_id);

  /* cleanup close MPI */

  if (yac_instance_count == 0) cleanup();
}

void yac_cfinalize () {

  if (default_instance_id != INT_MAX)
    yac_ccleanup_instance_(default_instance_id);
  default_instance_id = INT_MAX;

  /* cleanup close MPI */

  if (yac_instance_count == 0) cleanup();
}

/* ---------------------------------------------------------------------- */

void yac_cdef_datetime_instance (
  int yac_instance_id, const char * start_datetime,
  const char * end_datetime) {

  yac_instance_def_datetime(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"),
    start_datetime, end_datetime);
}

void yac_cdef_datetime ( const char * start_datetime,
                         const char * end_datetime ) {

  check_default_instance_id("yac_cdef_datetime");
  yac_cdef_datetime_instance(
    default_instance_id, start_datetime, end_datetime);
}

/* ---------------------------------------------------------------------- */

char * yac_cget_start_datetime_instance(int yac_instance_id) {

  return
    yac_instance_get_start_datetime(
      yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"));
}

char * yac_cget_start_datetime () {

  check_default_instance_id("yac_cget_start_datetime");
  return yac_cget_start_datetime_instance(default_instance_id);
}

/* ---------------------------------------------------------------------- */

char * yac_cget_end_datetime_instance(int yac_instance_id) {

  return
    yac_instance_get_end_datetime(
      yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"));
}

char * yac_cget_end_datetime () {

  check_default_instance_id("yac_cget_end_datetime");
  return yac_cget_end_datetime_instance(default_instance_id);
}

/* ---------------------------------------------------------------------- */

/* The YAC version number is provided in version.h
 */

char * yac_cget_version () {
  return yac_version;
}

/* ---------------------------------------------------------------------- */

void yac_cget_groupcomm_instance(int yac_instance_id, MPI_Comm * group_comm) {

  *group_comm =
    yac_instance_get_group_comm(
      yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"));
}

void yac_cget_groupcomm(MPI_Comm * group_comm) {

  check_default_instance_id("yac_cget_group_comm");
  yac_cget_groupcomm_instance(default_instance_id, group_comm);
}

/* internal routine to serve the Fortran interface */

void yac_cget_groupcomm_instance_c2f(int yac_instance_id,
                                     MPI_Fint *group_comm_f) {

  MPI_Comm group_comm;
  yac_cget_groupcomm_instance(yac_instance_id, &group_comm);

  *group_comm_f = MPI_Comm_c2f(group_comm);
}

void yac_cget_groupcomm_c2f(MPI_Fint *group_comm_f) {

  MPI_Comm group_comm;
  yac_cget_groupcomm(&group_comm);

  *group_comm_f = MPI_Comm_c2f(group_comm);
}

/* ---------------------------------------------------------------------- */

/* c interface routine */

void yac_cget_localcomm ( MPI_Comm *comm, int comp_id ) {

  struct user_input_data_component * comp_info =
    yac_unique_id_to_pointer(comp_id, "comp_id");

  MPI_Comm_dup(
    yac_component_get_communicator(
      yac_instance_get_component(
        comp_info->instance, comp_info->component_index)), comm);
}

/* internal routine to serve the Fortran interface */

void yac_get_localcomm_f2c ( MPI_Fint *comm_f, int comp_id ) {

  MPI_Comm comm;
  yac_cget_localcomm(&comm, comp_id);

  *comm_f = MPI_Comm_c2f(comm);
}

/* ---------------------------------------------------------------------- */

void yac_cset_comp_io_proc_step ( int comp_id, int io_proc_step ) {

  struct user_input_data_component * comp_info =
    yac_unique_id_to_pointer(comp_id, "comp_id");

  yac_component_set_io_proc_step(
    yac_instance_get_component(
      comp_info->instance, comp_info->component_index), io_proc_step);
}

/* ---------------------------------------------------------------------- */

void yac_cget_comp_io_proc_step ( int comp_id, int * io_proc_step ) {

  struct user_input_data_component * comp_info =
    yac_unique_id_to_pointer(comp_id, "comp_id");

  *io_proc_step =
    yac_component_get_io_proc_step(
      yac_instance_get_component(
        comp_info->instance, comp_info->component_index));
}

/* ---------------------------------------------------------------------- */

/* c interface routine */

void yac_cget_pair_rootcomm_instance(
  int yac_instance_id, MPI_Comm * comm,
  const char * comp_name1, const char * comp_name2 ) {

  *comm =
    yac_instance_get_pair_rootcomm(
      yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"),
      comp_name1, comp_name2);
}

void yac_cget_pair_rootcomm ( MPI_Comm *comm,
                              const char * comp_name1,
                              const char * comp_name2 ) {

  check_default_instance_id("yac_cget_pair_rootcomm");
  yac_cget_pair_rootcomm_instance(
    default_instance_id, comm, comp_name1, comp_name2);
}

/* internal routine to serve the Fortran interface */

void yac_cget_pair_rootcomm_instance_f2c ( int yac_instance_id,
                                           MPI_Fint *comm_f,
                                           const char * comp_name1,
                                           const char * comp_name2 ) {

  MPI_Comm comm;
  yac_cget_pair_rootcomm_instance(
    yac_instance_id, &comm, comp_name1, comp_name2);
  *comm_f = MPI_Comm_c2f(comm);
}

void yac_cget_pair_rootcomm_f2c ( MPI_Fint *comm_f,
                                  const char * comp_name1,
                                  const char * comp_name2 ) {

  MPI_Comm comm;
  yac_cget_pair_rootcomm(&comm, comp_name1, comp_name2);
  *comm_f = MPI_Comm_c2f(comm);
}

/* ---------------------------------------------------------------------- */

void yac_cdef_comps_instance(
  int yac_instance_id, char const ** comp_names, int num_comps,
  int * comp_ids) {

  struct yac_instance * instance =
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id");

  yac_instance_def_components(instance, comp_names, (size_t)num_comps);

  components =
    xrealloc(
      components, (num_components + (size_t)num_comps) * sizeof(*components));

  for (int i = 0; i < num_comps; ++i, ++num_components) {
    components[num_components] = xmalloc(1 * sizeof(**components));
    components[num_components]->instance = instance;
    components[num_components]->component_index = (size_t)i;
    comp_ids[i] = yac_pointer_to_unique_id(components[num_components]);
  }
}

void yac_cdef_comps (
  char const ** comp_names, int num_comps, int * comp_ids ) {

  check_default_instance_id("yac_cdef_comps");
  yac_cdef_comps_instance(
    default_instance_id, comp_names, num_comps, comp_ids);
}

void yac_cdef_comp_instance(
  int yac_instance_id, char const * comp_name, int * comp_id ) {

  yac_cdef_comps_instance(yac_instance_id, &comp_name, 1, comp_id);
}

void yac_cdef_comp ( char const * comp_name, int * comp_id ) {

  yac_cdef_comps(&comp_name, 1, comp_id);
}

/* ---------------------------------------------------------------------- */

static int user_input_data_add_points(
  struct yac_basic_grid * grid, enum yac_location location,
  coordinate_pointer coordinates) {

  struct user_input_data_points * curr_points = xmalloc(1 * sizeof(*curr_points));
  curr_points->default_mask_id = INT_MAX;
  curr_points->grid = grid;
  curr_points->location = location;
  curr_points->coordinates_idx =
    yac_basic_grid_add_coordinates_nocpy(grid, location, coordinates);

  points = xrealloc(points, (num_points + 1) * sizeof(*points));
  points[num_points] = curr_points;
  num_points++;

  return yac_pointer_to_unique_id(curr_points);
}

/* ---------------------------------------------------------------------- */

void yac_cdef_points_reg2d ( int const grid_id,
                             int const *nbr_points,
                             int const located,
                             double const *x_points,
                             double const *y_points,
                             int *point_id ) {

  enum yac_location location = yac_get_location(located);
  struct yac_basic_grid * grid = yac_unique_id_to_pointer(grid_id, "grid_id");

  size_t num_points = yac_basic_grid_get_data_size(grid, location);
  YAC_ASSERT(
    ((size_t)nbr_points[0] * (size_t)nbr_points[1]) == num_points,
    "ERROR(yac_cdef_points_reg2d): nbr_points does not match with grid")

  coordinate_pointer coordinates =
    xmalloc(num_points * sizeof(*coordinates));
  for (int i = 0, k = 0; i < nbr_points[1]; ++i)
    for (int j = 0; j < nbr_points[0]; ++j, ++k)
      LLtoXYZ(x_points[j], y_points[i], coordinates[k]);

  *point_id = user_input_data_add_points(grid, location, coordinates);
}

/* ---------------------------------------------------------------------- */

void yac_cdef_points_unstruct ( int const grid_id,
                                int const nbr_points,
                                int const located,
                                double const *x_points,
                                double const *y_points,
                                int *point_id ) {

  enum yac_location location = yac_get_location(located);
  struct yac_basic_grid * grid = yac_unique_id_to_pointer(grid_id, "grid_id");

  YAC_ASSERT(
    (size_t)nbr_points == yac_basic_grid_get_data_size(grid, location),
    "ERROR(yac_cdef_points_unstruct): nbr_points does not match with grid")

  coordinate_pointer coordinates =
    xmalloc((size_t)nbr_points * sizeof(*coordinates));
  for (int i = 0; i < nbr_points; ++i)
    LLtoXYZ(x_points[i], y_points[i], coordinates[i]);

  *point_id = user_input_data_add_points(grid, location, coordinates);
}

/* ---------------------------------------------------------------------- */

static int user_input_data_add_mask(
  struct yac_basic_grid * grid, enum yac_location location,
  int const * is_valid, size_t nbr_points) {

  struct user_input_data_masks * curr_mask = xmalloc(1 * sizeof(*curr_mask));
  curr_mask->grid = grid;
  curr_mask->location = location;
  curr_mask->masks_idx =
    yac_basic_grid_add_mask(grid, location, is_valid, nbr_points);

  masks = xrealloc(masks, (num_masks + 1) * sizeof(*masks));
  masks[num_masks] = curr_mask;
  num_masks++;

  return yac_pointer_to_unique_id(curr_mask);
}

/* ---------------------------------------------------------------------- */

void yac_cdef_mask ( int const grid_id,
                     int const nbr_points,
                     int const located,
                     int const * is_valid,
                     int *mask_id ) {

  enum yac_location location = yac_get_location(located);
  struct yac_basic_grid * grid = yac_unique_id_to_pointer(grid_id, "grid_id");

  YAC_ASSERT(
    (size_t)nbr_points == yac_basic_grid_get_data_size(grid, location),
    "ERROR(yac_cdef_mask): nbr_points does not match with grid")

  *mask_id =
    user_input_data_add_mask(grid, location, is_valid, (size_t)nbr_points);
}

void yac_cset_mask ( int const * is_valid, int points_id ) {

  struct user_input_data_points * points =
    yac_unique_id_to_pointer(points_id, "points_id");

  YAC_ASSERT(
    points->default_mask_id == INT_MAX,
    "ERROR(yac_cset_mask): default mask has already been set before")

  struct yac_basic_grid * grid = points->grid;
  enum yac_location location = points->location;
  size_t nbr_points = yac_basic_grid_get_data_size(grid, location);

  points->default_mask_id =
    user_input_data_add_mask(grid, location, is_valid, nbr_points);
}

/* ---------------------------------------------------------------------- */

void yac_cdef_field_mask ( char const * name,
                           int const comp_id,
                           int const * point_ids,
                           int const * mask_ids,
                           int const num_pointsets,
                           int * field_id ) {

  YAC_ASSERT(
    num_pointsets >= 1,
    "ERROR(yac_cdef_field_mask): invalid number of pointsets")

  YAC_ASSERT(
    point_ids != NULL,"ERROR(yac_cdef_field_mask): no point_ids provided")

  struct yac_basic_grid * grid =
    ((struct user_input_data_points *)
      yac_unique_id_to_pointer(point_ids[0], "point_ids[0]"))->grid;

  struct interp_field * interp_fields =
    xmalloc((size_t)num_pointsets * sizeof(*interp_fields));

  for (int i = 0; i < num_pointsets; ++i) {

    struct user_input_data_points * points =
      yac_unique_id_to_pointer(point_ids[i], "point_ids[i]");

    YAC_ASSERT(
      grid == points->grid,
      "ERROR(yac_cdef_field_mask): grid of point_ids do not match")

    size_t masks_idx;
    if (mask_ids[i] != INT_MAX) {

      struct user_input_data_masks * mask =
        yac_unique_id_to_pointer(mask_ids[i], "mask_ids[i]");

      YAC_ASSERT(
        mask->grid == points->grid,
        "ERROR(yac_cdef_field_mask): grids of mask and points do not match")
      YAC_ASSERT(
        mask->location == points->location,
        "ERROR(yac_cdef_field_mask): location of mask and points do not match")
      masks_idx = mask->masks_idx;
    } else {
      masks_idx = SIZE_MAX;
    }

    interp_fields[i].location = points->location;
    interp_fields[i].coordinates_idx = points->coordinates_idx;
    interp_fields[i].masks_idx = masks_idx;
  }

  struct user_input_data_component * comp_info =
    yac_unique_id_to_pointer(comp_id, "comp_id");

  *field_id =
    yac_pointer_to_unique_id(
      yac_instance_add_field(
        comp_info->instance, name, comp_info->component_index,
        grid, interp_fields, num_pointsets));

  free(interp_fields);
}

void yac_cdef_field ( char const * name,
                      int const comp_id,
                      int const * point_ids,
                      int const num_pointsets,
                      int * field_id ) {

  YAC_ASSERT(
    num_pointsets >= 1, "ERROR(yac_cdef_field): invalid number of pointsets")

  YAC_ASSERT(point_ids != NULL, "ERROR(yac_cdef_field): no point_ids provided")

  int * mask_ids = xmalloc((size_t)num_pointsets * sizeof(*mask_ids));

  for (int i = 0; i < num_pointsets; ++i)
    mask_ids[i] =
      ((struct user_input_data_points *)
         yac_unique_id_to_pointer(point_ids[i], "point_ids[i]"))->
           default_mask_id;

  yac_cdef_field_mask (
    name, comp_id, point_ids, mask_ids, num_pointsets, field_id );

  free(mask_ids);
}

/* ---------------------------------------------------------------------- */

static double ** get_recv_field_pointers(
  int const field_id,
  int const collection_size,
  double *recv_field) { // recv_field[collection_size][Points]

  struct coupling_field * cpl_field =
    yac_unique_id_to_pointer(field_id, "field_id");

  YAC_ASSERT(
    yac_coupling_field_get_num_interp_fields(cpl_field) == 1,
    "ERROR(get_recv_field_pointers): invalid number of interpolation fields "
    "(should be one)")

  enum yac_location location =
    yac_coupling_field_get_interp_fields(cpl_field)->location;
  size_t num_points =
    yac_coupling_field_get_data_size(cpl_field, location);

  double ** recv_field_ = xmalloc(collection_size * sizeof(*recv_field_));

  for (int i = 0; i < collection_size; ++i) {
    recv_field_[i] = recv_field;
    recv_field += num_points;
  }

  return recv_field_;
}

void yac_cget_ ( int const field_id,
                 int const collection_size,
                 double *recv_field,   // recv_field[collection_size][Points]
                 int    *info,
                 int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */
  double ** recv_field_ =
    get_recv_field_pointers(field_id, collection_size, recv_field);

  yac_cget ( field_id, collection_size, recv_field_, info, ierr );

  free(recv_field_);
}

/* ---------------------------------------------------------------------- */

static struct interpolation * yac_cget_pre_processing(
  int const field_id, int collection_size, int *info, int *ierr) {

  *ierr = -1;
  *info = NONE;

  struct coupling_field * cpl_field =
    yac_unique_id_to_pointer(field_id, "field_id");

  if (yac_get_coupling_field_exchange_type(cpl_field) != TARGET)
    return NULL;

  /* --------------------------------------------------------------------
     Check for restart and coupling events
     -------------------------------------------------------------------- */

  struct event * event = yac_get_coupling_field_get_op_event(cpl_field);

  int action = yac_event_check ( event );

  /* --------------------------------------------------------------------
     add one model time step (as provided in coupling.xml) to the current event date
     -------------------------------------------------------------------- */

  yac_event_update ( event );

  /* ------------------------------------------------------------------
     return in case we are already beyond the end of the run
     ------------------------------------------------------------------ */

  if ( action == OUT_OF_BOUND ) {

      fprintf (stderr, "WARNING: yac get action is beyond end of run date!\n");
      fprintf (stdout, "WARNING: yac get action is beyond end of run date!\n");

      *info = OUT_OF_BOUND;
      *ierr = 0;

      return  NULL;
  }

  /* --------------------------------------------------------------------
     start actions
     -------------------------------------------------------------------- */

  if ( action == NONE ) {
    *info = action;
    *ierr = 0;

    return NULL;
  }

  YAC_ASSERT(
    collection_size == yac_get_coupling_field_collection_size(cpl_field),
    "ERROR: collection size does not match with xml.")

  if ( action == RESTART ) action = GET_FOR_RESTART;

  *info = action;
  *ierr = 0;

  return yac_get_coupling_field_get_op_interpolation(cpl_field);
}

void yac_cget ( int const field_id,
                int collection_size,
                double ** recv_field, // recv_field[collection_size][Points]
                int    *info,
                int    *ierr ) {

  struct interpolation * interpolation =
    yac_cget_pre_processing(field_id, collection_size, info, ierr);

  if ((*ierr == 0) && (interpolation != NULL))
    yac_interpolation_execute_get(interpolation, recv_field);
}

/* ---------------------------------------------------------------------- */

static double *** get_send_field_pointers(
  int const field_id,
  int const collection_size,
  double *send_field) { // send_field[collection_size][nPointSets][Points]

  struct coupling_field * cpl_field =
    yac_unique_id_to_pointer(field_id, "field_id");

  size_t num_interp_fields =
    yac_coupling_field_get_num_interp_fields(cpl_field);

  size_t * num_points = xmalloc(num_interp_fields * sizeof(*num_points));

  for (size_t i = 0; i < num_interp_fields; i++) {

    enum yac_location location =
      yac_coupling_field_get_interp_fields(cpl_field)[i].location;
    num_points[i] = yac_coupling_field_get_data_size(cpl_field, location);
  }

  double *** send_field_ = xmalloc(collection_size * sizeof(*send_field_));

  for (int i = 0; i < collection_size; ++i) {

    send_field_[i] = xmalloc(num_interp_fields * sizeof(**send_field_));

    for (size_t j = 0; j < num_interp_fields; ++j) {

      send_field_[i][j] = send_field;
      send_field += num_points[j];
    }
  }
  free(num_points);

  return send_field_;
}

void yac_cput_ ( int const field_id,
                 int const collection_size,
                 double *send_field,   // send_field[collection_size]
                                       //           [nPointSets]
                                       //           [Points]
                 int    *info,
                 int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */
  double *** send_field_ =
    get_send_field_pointers(field_id, collection_size, send_field);

  yac_cput ( field_id, collection_size, send_field_, info, ierr );

  for (int i = 0; i < collection_size; ++i) free(send_field_[i]);
  free(send_field_);
}

/* ---------------------------------------------------------------------- */

static double *** get_send_field_pointers_ptr_(
  int const field_id,
  int const collection_size,
  double **send_field) { // send_field[collection_size*nPointSets][Points]

  struct coupling_field * cpl_field =
    yac_unique_id_to_pointer(field_id, "field_id");

  size_t num_interp_fields =
    yac_coupling_field_get_num_interp_fields(cpl_field);

  size_t * num_points = xmalloc(num_interp_fields * sizeof(*num_points));

  for (size_t i = 0; i < num_interp_fields; i++) {

    enum yac_location location =
      yac_coupling_field_get_interp_fields(cpl_field)[i].location;
    num_points[i] = yac_coupling_field_get_data_size(cpl_field, location);
  }

  double *** send_field_ = xmalloc(collection_size * sizeof(*send_field_));

  for (int i = 0, k = 0; i < collection_size; ++i) {

    send_field_[i] = xmalloc(num_interp_fields * sizeof(**send_field_));
    for (size_t j = 0; j < num_interp_fields; ++j, ++k)
      send_field_[i][j] = send_field[k];
  }
  free(num_points);

  return send_field_;
}

void yac_cput_ptr_ ( int const field_id,
                     int const collection_size,
                     double ** send_field,   // send_field[collection_size *
                                             //            nPointSets][Points]
                     int    *info,
                     int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */
  double *** send_field_ =
    get_send_field_pointers_ptr_(field_id, collection_size, send_field);

  yac_cput ( field_id, collection_size, send_field_, info, ierr );

  for (int i = 0; i < collection_size; ++i) free(send_field_[i]);
  free(send_field_);
}

/* ---------------------------------------------------------------------- */

static struct interpolation * yac_cput_pre_processing(
  struct coupling_field * cpl_field, unsigned put_idx,
  int const collection_size, double *** send_field,
  double **** send_field_acc_, int *info, int *ierr) {

  *ierr = 0;

  /* ------------------------------------------------------------------
     Check for restart and coupling events
     ------------------------------------------------------------------ */
  struct event * event = yac_get_coupling_field_put_op_event(cpl_field, put_idx);
  int action = yac_event_check(event);
  *info = action;

  /* ------------------------------------------------------------------
     add one model time step (as provided in coupling.xml) to the current event date
     ------------------------------------------------------------------ */
  yac_event_update(event);

  /* ------------------------------------------------------------------
     return in case we are already beyond the end of the run
     ------------------------------------------------------------------ */
  if (action == OUT_OF_BOUND) {

    fprintf(stderr, "WARNING: yac put action is beyond end of run date!\n");
    fprintf(stdout, "WARNING: yac put action is beyond end of run date!\n");
    *send_field_acc_ = NULL;
    return NULL;
  }

  /* ------------------------------------------------------------------
     start actions
     ------------------------------------------------------------------ */
  int time_operation = yac_get_event_time_operation(event);
  if ( action == NONE && time_operation == INSTANT ) return NULL;

  /* ------------------------------------------------------------------
     If it is time for restart, set appropriate flags and continue
     ------------------------------------------------------------------ */

  if ( action == RESTART ) *info = PUT_FOR_RESTART;

  /* ------------------------------------------------------------------
     First deal with instant sends
     ------------------------------------------------------------------ */

  if ( time_operation == INSTANT ) {
    *info = MAX(COUPLING, *info);
    *send_field_acc_ = NULL;
    return yac_get_coupling_field_put_op_interpolation(cpl_field, put_idx);
  }

  /* ------------------------------------------------------------------
     Accumulation & Averaging
     ------------------------------------------------------------------ */

  /* accumulate data */

  double *** send_field_acc =
    yac_get_coupling_field_put_op_send_field_acc(cpl_field, put_idx);
  *send_field_acc_ = send_field_acc;

  int time_accumulation_count =
    yac_get_coupling_field_put_op_time_accumulation_count(cpl_field, put_idx);
  time_accumulation_count++;

  size_t num_interp_fields =
    yac_coupling_field_get_num_interp_fields(cpl_field);

  for (int h = 0; h < collection_size; h++) {
    for (size_t i = 0; i < num_interp_fields; i++) {
      size_t num_points =
        yac_coupling_field_get_data_size(
          cpl_field,
          yac_coupling_field_get_interp_fields(cpl_field)[i].location);
      for (size_t j = 0; j < num_points; ++j)
        send_field_acc[h][i][j] += send_field[h][i][j];
    }
  }

/* --------------------------------------------------------------------
   Check whether we have to perform the coupling in this call
   -------------------------------------------------------------------- */

  if (action == NONE) {
    yac_set_coupling_field_put_op_time_accumulation_count(
      cpl_field, put_idx, time_accumulation_count);
    return NULL;
  }

/* --------------------------------------------------------------------
   Average data if required
   -------------------------------------------------------------------- */

  if ( time_operation == TIME_AVERAGE ) {
    double weight = 1.0 / (double)time_accumulation_count;

    for (int h = 0; h < collection_size; ++h) {
      for (size_t i = 0; i < num_interp_fields; i++) {
        size_t num_points =
          yac_coupling_field_get_data_size(
            cpl_field,
            yac_coupling_field_get_interp_fields(cpl_field)[i].location);
        for (size_t j = 0; j < num_points; j++) {
          send_field_acc[h][i][j] *= weight;
        }
      }
    }
  }

/* --------------------------------------------------------------------
   return interpolation
   -------------------------------------------------------------------- */

  *info = MAX(COUPLING, *info);
  return yac_get_coupling_field_put_op_interpolation(cpl_field, put_idx);
}

void yac_cput ( int const field_id,
                int const collection_size,
                double *** const send_field,   // send_field[collection_size]
                                               //           [nPointSets][Points]
                int    *info,
                int    *ierr ) {

  *info = NONE;
  *ierr = -1;

  struct coupling_field * cpl_field =
    yac_unique_id_to_pointer(field_id, "field_id");

  if (yac_get_coupling_field_exchange_type(cpl_field) != SOURCE)
    return;

  YAC_ASSERT(
    collection_size == yac_get_coupling_field_collection_size(cpl_field),
    "ERROR: collection size does not match with xml.")

  for (unsigned put_idx = 0;
       put_idx < yac_get_coupling_field_num_puts(cpl_field); ++put_idx) {

    int curr_action;
    int curr_ierr;
    double *** send_field_acc;
    struct interpolation * interpolation =
      yac_cput_pre_processing(
        cpl_field, put_idx, collection_size, send_field,
        &send_field_acc, &curr_action, &curr_ierr);

    *info = MAX(*info, curr_action);
    *ierr = MAX(*ierr, curr_ierr);

    /* ------------------------------------------------------------------
       return in case we are already beyond the end of the run
       ------------------------------------------------------------------ */
    if (curr_action == OUT_OF_BOUND) {
      *info = OUT_OF_BOUND;
      return;
    }

    /* ------------------------------------------------------------------
       in case there is nothing to be done for the current put
       ------------------------------------------------------------------ */
    if (curr_action == NONE) continue;

    /* ------------------------------------------------------------------
       in case we are supposed to couple
       ------------------------------------------------------------------ */
    if ((*ierr == 0) && (interpolation != NULL)) {
      double *** send_field_ptr =
        (send_field_acc == NULL)?send_field:send_field_acc;
      yac_interpolation_execute_put(interpolation, send_field_ptr);

      if (send_field_acc != NULL) {

        /* --------------------------------------------------------------------
           Initialise memory for the next accumulation
           -------------------------------------------------------------------- */
        yac_init_coupling_field_put_op_send_field_acc(cpl_field, put_idx, 0.0);
        yac_set_coupling_field_put_op_time_accumulation_count(
          cpl_field, put_idx, 0);
      }
    }
  }
}

/* ---------------------------------------------------------------------- */


void yac_cexchange_ ( int const send_field_id,
                      int const recv_field_id,
                      int const collection_size,
                      double *send_field, // send_field[collection_size]
                                          //           [nPointSets]
                                          //           [Points]
                      double *recv_field, // recv_field[collection_size][Points]
                      int    *send_info,
                      int    *recv_info,
                      int    *ierr ) {

  /* Needed to transfer from Fortran data structure to C */
  double *** send_field_ =
    get_send_field_pointers(send_field_id, collection_size, send_field);
  double ** recv_field_ =
    get_recv_field_pointers(recv_field_id, collection_size, recv_field);

  yac_cexchange(
    send_field_id, recv_field_id, collection_size, send_field_, recv_field_,
    send_info, recv_info, ierr);

  free(recv_field_);
  for (int i = 0; i < collection_size; ++i) free(send_field_[i]);
  free(send_field_);
}

/* ---------------------------------------------------------------------- */


void yac_cexchange_ptr_ ( int const send_field_id,
                          int const recv_field_id,
                          int const collection_size,
                          double ** send_field, // send_field[collection_size *
                                                //            nPointSets][Points]
                          double ** recv_field, // recv_field[collection_size][Points]
                          int    *send_info,
                          int    *recv_info,
                          int    *ierr ) {


  /* Needed to transfer from Fortran data structure to C */
  double *** send_field_ =
    get_send_field_pointers_ptr_(send_field_id, collection_size, send_field);

  yac_cexchange(
    send_field_id, recv_field_id, collection_size, send_field_, recv_field,
    send_info, recv_info, ierr);

  for (int i = 0; i < collection_size; ++i) free(send_field_[i]);
  free(send_field_);

}

/* ---------------------------------------------------------------------- */

void yac_cexchange ( int const send_field_id,
                     int const recv_field_id,
                     int const collection_size,
                     double *** const send_field, // send_field[collection_size]
                                                  //           [nPointSets][Points]
                     double ** recv_field, // recv_field[collection_size][Points]
                     int    *send_info,
                     int    *recv_info,
                     int    *ierr ) {

  *send_info = NONE;
  *ierr = -1;

  struct coupling_field * send_cpl_field =
    yac_unique_id_to_pointer(send_field_id, "send_field_id");

  if (yac_get_coupling_field_exchange_type(send_cpl_field) != SOURCE) {
    yac_cget(recv_field_id, collection_size, recv_field, recv_info, ierr);
    return;
  }

  YAC_ASSERT(
    collection_size == yac_get_coupling_field_collection_size(send_cpl_field),
    "ERROR(yac_cexchange): collection size does not match with xml.")

  YAC_ASSERT(
    yac_get_coupling_field_num_puts(send_cpl_field) == 1,
    "ERROR(yac_cexchange): more than one put per field is not supported "
    "for yac_cexchange.")

  int send_action;
  int send_ierr;
  double *** send_field_acc;
  struct interpolation * put_interpolation =
    yac_cput_pre_processing(
      send_cpl_field, 0, collection_size, send_field,
      &send_field_acc, &send_action, &send_ierr);

  *send_info = MAX(*send_info, send_action);
  *ierr = MAX(*ierr, send_ierr);

  /* ------------------------------------------------------------------
     return in case we are already beyond the end of the run for the puts
     or there is nothing to be done for the current put
     ------------------------------------------------------------------ */
  if ((send_action == OUT_OF_BOUND) || (send_action == NONE)) {
    yac_cget(recv_field_id, collection_size, recv_field, recv_info, ierr);
    return;
  }

  /* ------------------------------------------------------------------
     check the get
     ------------------------------------------------------------------ */

  struct interpolation * get_interpolation =
    yac_cget_pre_processing(recv_field_id, collection_size, recv_info, ierr);

  /* ------------------------------------------------------------------
     do the required exchanges
     ------------------------------------------------------------------ */

  if (*ierr == 0) {

    double *** send_field_ptr =
      (send_field_acc == NULL)?send_field:send_field_acc;

    // if the get is active
    if (get_interpolation != NULL) {

      YAC_ASSERT(
        get_interpolation == put_interpolation,
        "ERROR(yac_cexchange): send_field_id and recv_field_id do not match")

      yac_interpolation_execute(
        put_interpolation, send_field_ptr, recv_field);
    } else {

      // just execute the put
      yac_interpolation_execute_put(put_interpolation, send_field_ptr);
    }

    if (send_field_acc != NULL) {

      /* --------------------------------------------------------------------
         Initialise memory for the next accumulation
         -------------------------------------------------------------------- */
      yac_init_coupling_field_put_op_send_field_acc(send_cpl_field, 0, 0.0);
      yac_set_coupling_field_put_op_time_accumulation_count(
        send_cpl_field, 0, 0);
    }
  }
}

/* ---------------------------------------------------------------------- */

void yac_csearch_instance(int yac_instance_id, int * ierror) {

  yac_setup_instance(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"));

  *ierror = 0;
}

void yac_csearch ( int * ierror ) {

  check_default_instance_id("yac_csearch");
  yac_csearch_instance(default_instance_id, ierror);
}

/* ----------------------------------------------------------------------
                   query functions
   ----------------------------------------------------------------------*/

void yac_cget_nbr_fields_instance(int yac_instance_id, int *nbr_fields) {

  struct coupling_field ** fields;
  size_t num_fields;
  yac_instance_get_fields(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"),
    &fields, &num_fields);

  *nbr_fields = (int)num_fields;
}

void yac_cget_nbr_fields ( int *nbr_fields ) {

  check_default_instance_id("yac_cget_nbr_fields");
  yac_cget_nbr_fields_instance(default_instance_id, nbr_fields);
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_ids_instance(
  int yac_instance_id, int nbr_fields, int *field_ids) {

  struct coupling_field ** fields;
  size_t num_fields;
  yac_instance_get_fields(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"),
    &fields, &num_fields);

  YAC_ASSERT(
    nbr_fields == (int)num_fields,
    "ERROR(yac_cget_field_ids_instance): nbr of fields does not match.")

  for ( size_t i = 0; i < num_fields; ++i ) {
    int temp_field_id = yac_lookup_pointer(fields[i]);
    YAC_ASSERT(
      temp_field_id != UINT_MAX,
      "ERROR(yac_cget_field_ids_instance): invalid field id")
    field_ids[i] = (int)temp_field_id;
  }
}

void yac_cget_field_ids ( int nbr_fields, int *field_ids ) {

  check_default_instance_id("yac_cget_field_ids");
  yac_cget_field_ids_instance(default_instance_id, nbr_fields, field_ids);
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_id_from_field_and_grid_name_instance(
  int yac_instance_id, const char * field_name, const char * grid_name,
  int *field_id) {

  struct coupling_field ** fields;
  size_t num_fields;
  yac_instance_get_fields(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"),
    &fields, &num_fields);

  int field_id_ = -1;

  for ( size_t i = 0; i < num_fields; ++i ) {
    if (!strcmp(yac_get_coupling_field_name(fields[i]),field_name) &&
        !strcmp(
          grid_name,
          yac_basic_grid_get_name(
            yac_coupling_field_get_basic_grid(fields[i])))) {
      YAC_ASSERT_F(
        field_id_ == -1,
        "ERROR(yac_cget_field_id_from_field_and_grid_name_instance): "
        "field \"%s\" define more than once for grid \"%s\"",
        field_name, grid_name);
      int temp_field_id = yac_lookup_pointer(fields[i]);
      YAC_ASSERT(
        temp_field_id != UINT_MAX,
        "ERROR(yac_cget_field_id_from_field_and_grid_name_instance): "
        "invalid field id")
      field_id_ = (int)temp_field_id;
    }
  }

  YAC_ASSERT_F(
    field_id_ != -1,
    "ERROR(yac_cget_field_id_from_field_and_grid_name_instance): "
    "no field define with name \"%s\" on grid \"%s\"", field_name, grid_name)
  *field_id = field_id_;
}

void yac_cget_field_id_from_field_and_grid_name(
  const char * field_name, const char * grid_name, int *field_id) {

  check_default_instance_id("yac_cget_field_id_from_field_and_grid_name");
  yac_cget_field_id_from_field_and_grid_name_instance(
    default_instance_id, field_name, grid_name, field_id);
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_id_from_name_instance(
  int yac_instance_id, const char *field_name, int *field_id) {

  struct coupling_field ** fields;
  size_t num_fields;
  yac_instance_get_fields(
    yac_unique_id_to_pointer(yac_instance_id, "yac_instance_id"),
    &fields, &num_fields);

  int field_id_ = -1;

  for ( size_t i = 0; i < num_fields; ++i ) {
    if (!strcmp(yac_get_coupling_field_name(fields[i]),field_name)) {
      YAC_ASSERT_F(
        field_id_ == -1,
        "ERROR(yac_cget_field_id_from_name_instance): "
        "field \"%s\" define more than once, use "
        "yac_cget_field_from_field_and_grid_name_instance to retrieve field_id",
        field_name);
      int temp_field_id = yac_lookup_pointer(fields[i]);
      YAC_ASSERT(
        temp_field_id != UINT_MAX,
        "ERROR(yac_cget_field_id_from_name_instance): invalid field id")
      field_id_ = (int)temp_field_id;
    }
  }

  YAC_ASSERT_F(
    field_id_ != -1, "ERROR(yac_cget_field_id_from_name_instance): "
    "no field define with name \"%s\"", field_name)
  *field_id = field_id_;
}

void yac_cget_field_id_from_name ( const char *field_name,
                                   int *field_id ) {

  check_default_instance_id("yac_cget_field_id_from_name");
  yac_cget_field_id_from_name_instance(
    default_instance_id, field_name, field_id);
}

/* ---------------------------------------------------------------------- */

void yac_cget_field_name_from_id ( int field_id,
                                   char *field_name ) {

  struct coupling_field * field =
    yac_unique_id_to_pointer(field_id, "field_id");

  YAC_ASSERT(field != NULL, "ERROR: field ID not defined!")

  strcpy(field_name, yac_get_coupling_field_name(field));
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_id ( int field_id, int *role ) {

  struct coupling_field * field =
    yac_unique_id_to_pointer(field_id, "field_id");

  YAC_ASSERT(field != NULL, "ERROR: field ID not defined!");

  *role = yac_get_coupling_field_exchange_type(field);
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_name_instance(
  int yac_instance_id, char const *field_name, int *role) {

  int field_id;
  yac_cget_field_id_from_name_instance(
    yac_instance_id, field_name, &field_id);

  *role =
    yac_get_coupling_field_exchange_type(
      yac_unique_id_to_pointer(field_id, "field_id"));
}

void yac_cget_role_of_field_by_name ( char const *field_name, int *role ) {

  check_default_instance_id("yac_cget_role_of_field_by_name");
  yac_cget_role_of_field_by_name_instance
    (default_instance_id, field_name, role);
}

/* ---------------------------------------------------------------------- */

void yac_cget_role_of_field_by_field_and_grid_name_instance(
  int yac_instance_id, char const *field_name, char const *grid_name,
  int *role) {

  int field_id;
  yac_cget_field_id_from_field_and_grid_name_instance(
    yac_instance_id, field_name, grid_name, &field_id);

  *role =
    yac_get_coupling_field_exchange_type(
      yac_unique_id_to_pointer(field_id, "field_id"));
}

void yac_cget_role_of_field_by_field_and_grid_name(
  char const *field_name, char const *grid_name, int *role ) {

  check_default_instance_id("yac_cget_role_of_field_by_field_and_grid_name");
  yac_cget_role_of_field_by_field_and_grid_name_instance
    (default_instance_id, field_name, grid_name, role);
}

/* ---------------------------------------------------------------------- */

void yac_cget_coupling_timestep_name_instance ( int yac_instance_id,
                                                const char * field_name,
                                                char *timestep_string ) {

  int field_id;
  yac_cget_field_id_from_name_instance(
    yac_instance_id, field_name, &field_id);
  yac_cget_coupling_timestep_id(field_id, timestep_string);
}

void yac_cget_coupling_timestep_name ( const char * field_name,
                                       char *timestep_string ) {

  check_default_instance_id("yac_cget_coupling_timestep_name");
  yac_cget_coupling_timestep_name_instance(
    default_instance_id, field_name, timestep_string);
}

/* ---------------------------------------------------------------------- */

void yac_cget_coupling_timestep_id ( int field_id,
                                     char *timestep_string ) {

  struct coupling_field * field =
    yac_unique_id_to_pointer(field_id, "field_id");
  struct event * event = NULL;

  YAC_ASSERT(field != NULL, "ERROR: field ID not defined!")

  int role = yac_get_coupling_field_exchange_type(field);

  if ( role == SOURCE )
    for (unsigned put_idx = 0;
         put_idx < yac_get_coupling_field_num_puts(field); ++put_idx)
      event = yac_get_coupling_field_put_op_event(field, put_idx);

  else if ( role == TARGET )
    event = yac_get_coupling_field_get_op_event(field);

  YAC_ASSERT(event != NULL, "ERROR: event not defined!")

  char my_string[MAX_TIMEDELTA_STR_LEN];

  if ( role == SOURCE || role == TARGET )
    strcpy(timestep_string, yac_get_event_coupling_timestep(event, my_string));
}

/* ---------------------------------------------------------------------- */

void yac_cget_collection_size_name_instance ( int yac_instance_id,
                                              const char * field_name,
                                              int *collection_size ) {

  int field_id;
  yac_cget_field_id_from_name_instance(
    yac_instance_id, field_name, &field_id);
  yac_cget_collection_size_id(field_id, collection_size);
}

void yac_cget_collection_size_name ( const char * field_name,
                                     int *collection_size ) {

  check_default_instance_id("yac_cget_collection_size_name");
  yac_cget_collection_size_name_instance(
    default_instance_id, field_name, collection_size);
}

/* ---------------------------------------------------------------------- */

void yac_cget_collection_size_id ( int field_id,
                                   int *collection_size ) {

  struct coupling_field * field =
    yac_unique_id_to_pointer(field_id, "field_id");

  *collection_size = (int)yac_get_coupling_field_collection_size(field);
}

/* ---------------------------------------------------------------------- */

void yac_cget_model_timestep_name_instance ( int yac_instance_id,
                                             const char * field_name,
                                             char *timestep_string ) {

  int field_id;
  yac_cget_field_id_from_name_instance(
    yac_instance_id, field_name, &field_id);
  yac_cget_model_timestep_id ( field_id, timestep_string );
}

void yac_cget_model_timestep_name ( const char * field_name,
                                    char *timestep_string ) {

  check_default_instance_id("yac_cget_model_timestep_name");
  yac_cget_model_timestep_name_instance(
    default_instance_id, field_name, timestep_string);
}

/* ---------------------------------------------------------------------- */

void yac_cget_model_timestep_id ( int field_id,
                                  char *timestep_string ) {

  struct coupling_field * field =
    yac_unique_id_to_pointer(field_id, "field_id");
  struct event * event = NULL;

  YAC_ASSERT(field != NULL, "ERROR: field ID not defined!")

  int role = yac_get_coupling_field_exchange_type(field);

  YAC_ASSERT(
    (role == SOURCE) || (role == TARGET),
    "ERROR(yac_cget_model_timestep_id): field is neither SOURCE nor TARGET")

  if ( role == SOURCE )
    for (unsigned put_idx = 0;
         put_idx < yac_get_coupling_field_num_puts(field); ++put_idx)
      event = yac_get_coupling_field_put_op_event(field, put_idx);

  else
    event = yac_get_coupling_field_get_op_event(field);

  YAC_ASSERT(event != NULL, "ERROR: event not defined!")

  char my_string[MAX_TIMEDELTA_STR_LEN];
  strcpy(timestep_string,yac_get_event_model_timestep(event, my_string));
}

/* ---------------------------------------------------------------------- */

void yac_cdef_grid_reg2d(
  const char * grid_name, int nbr_vertices[2], int cyclic[2],
  double *x_vertices, double *y_vertices, int *grid_id) {

  size_t nbr_vertices_size_t[2] =
    {(size_t)nbr_vertices[0], (size_t)nbr_vertices[1]};

  *grid_id =
    yac_add_grid(
      grid_name, yac_generate_basic_grid_data_reg_2d(
                   nbr_vertices_size_t, cyclic, x_vertices, y_vertices));
}

/* ---------------------------------------------------------------------- */

void yac_cdef_grid_unstruct(
  const char * grid_name, int nbr_vertices,
  int nbr_cells, int *num_vertices_per_cell, double *x_vertices,
  double *y_vertices, int *cell_to_vertex, int *grid_id) {

  *grid_id =
    yac_add_grid(
      grid_name,
      yac_generate_basic_grid_data_unstruct(
        (size_t)nbr_vertices, (size_t)nbr_cells, num_vertices_per_cell,
        x_vertices, y_vertices, cell_to_vertex));
}

/* ---------------------------------------------------------------------- */

void yac_cset_global_index(
  int const * global_index, int location, int grid_id) {

  struct basic_grid_data * grid_data =
    yac_basic_grid_get_data(yac_unique_id_to_pointer(grid_id, "grid_id"));

  yac_int ** grid_global_ids;
  size_t count;

  YAC_ASSERT(
    (location == CELL) || (location == CORNER) || (location == EDGE),
    "ERROR(yac_cset_global_index): invalid location")
  switch (location) {
    case (CELL): {
      grid_global_ids = &(grid_data->cell_ids);
      count = grid_data->num_cells;
      break;
    }
    case (CORNER): {
      grid_global_ids = &(grid_data->vertex_ids);
      count = grid_data->num_vertices;
      break;
    }
    default:
    case (EDGE): {
      grid_global_ids = &(grid_data->edge_ids);
      count = grid_data->num_edges;
      break;
    }
  }

  size_t global_ids_size = count * sizeof(**grid_global_ids);
  if (*grid_global_ids == NULL) *grid_global_ids = xmalloc(global_ids_size);
  for (size_t i = 0; i < count; ++i)
    (*grid_global_ids)[i] = (yac_int)(global_index[i]);
}

/* ---------------------------------------------------------------------- */

void yac_cset_core_mask(
  int const * is_core, int location, int grid_id) {

  struct basic_grid_data * grid_data =
    yac_basic_grid_get_data(yac_unique_id_to_pointer(grid_id, "grid_id"));

  int ** grid_core_mask;
  size_t count;

  YAC_ASSERT(
    (location == CELL) || (location == CORNER) || (location == EDGE),
    "ERROR(yac_cset_core_mask): invalid location")
  switch (location) {
    case (CELL): {
      grid_core_mask = &(grid_data->core_cell_mask);
      count = grid_data->num_cells;
      break;
    }
    case (CORNER): {
      grid_core_mask = &(grid_data->core_vertex_mask);
      count = grid_data->num_vertices;
      break;
    }
    default:
    case (EDGE): {
      grid_core_mask = &(grid_data->core_edge_mask);
      count = grid_data->num_edges;
      break;
    }
  }

  size_t core_mask_size = count * sizeof(**grid_core_mask);
  if (*grid_core_mask == NULL) *grid_core_mask = xmalloc(core_mask_size);
  memcpy(*grid_core_mask, is_core, core_mask_size);
}

/* ---------------------------------------------------------------------- */

size_t yac_get_grid_size ( int located, int grid_id ) {

  struct yac_basic_grid * grid =
    yac_unique_id_to_pointer(grid_id, "grid_id");
  enum yac_location location = yac_get_location(located);

  return yac_basic_grid_get_data_size(grid, location);
}

/* ---------------------------------------------------------------------- */

size_t yac_get_points_size ( int points_id  ) {

  struct user_input_data_points * points =
    yac_unique_id_to_pointer(points_id, "points_id");
  struct yac_basic_grid * grid = points->grid;
  enum yac_location location = points->location;

  return yac_basic_grid_get_data_size(grid, location);
}

/* ---------------------------------------------------------------------- */

// gets the number of active transients for a component
int yac_get_transients_count(int comp_id) {

  struct user_input_data_component * comp_info =
    yac_unique_id_to_pointer(comp_id, "comp_id");
  struct yac_instance * instance = comp_info->instance;
  struct yac_couple_config * couple_config =
    yac_instance_get_couple_config(instance);
  struct component * component =
    yac_instance_get_component(instance, comp_info->component_index);
  int global_comp_id = (int)yac_component_get_global_id(component);

  size_t num_couples = yac_couple_config_get_num_couples(couple_config);

  int transient_count = 0;

  for (size_t couple_idx = 0; couple_idx < num_couples; ++couple_idx) {

    int couple_comp_ids[2];
    yac_couple_config_get_couple_component_ids(
      couple_config, couple_idx, couple_comp_ids);

    if ((couple_comp_ids[0] == global_comp_id) ||
        (couple_comp_ids[1] == global_comp_id))
      transient_count +=
        yac_couple_config_get_num_couple_transients(couple_config, couple_idx);
  }
  return transient_count;
}

// gets ids for all active transients
void yac_get_transient_ids(int comp_id, int * transient_ids) {

  struct user_input_data_component * comp_info =
    yac_unique_id_to_pointer(comp_id, "comp_id");
  struct yac_instance * instance = comp_info->instance;
  struct yac_couple_config * couple_config =
    yac_instance_get_couple_config(instance);
  struct component * component =
    yac_instance_get_component(instance, comp_info->component_index);
  int global_comp_id = (int)yac_component_get_global_id(component);

  size_t num_couples = yac_couple_config_get_num_couples(couple_config);

  for (size_t couple_idx = 0, transient_idx = 0; couple_idx < num_couples;
      ++couple_idx) {

    int couple_comp_ids[2];
    yac_couple_config_get_couple_component_ids(
      couple_config, couple_idx, couple_comp_ids);

    if ((couple_comp_ids[0] != global_comp_id) &&
        (couple_comp_ids[1] != global_comp_id)) continue;

    size_t num_couple_transients =
      yac_couple_config_get_num_couple_transients(couple_config, couple_idx);

    for (size_t transient_couple_idx = 0;
        transient_couple_idx < num_couple_transients; ++transient_couple_idx) {

      struct couple_config_transient * transient =
        xmalloc(1 * sizeof(*transient));
      transient->global_comp_id = global_comp_id;
      transient->couple_config = couple_config;
      transient->couple_idx = couple_idx;
      transient->transient_couple_idx = transient_couple_idx;

      transients =
        xrealloc(transients, (num_transients + 1) * sizeof(*transients));
      transients[num_transients] = transient;
      num_transients++;

      transient_ids[transient_idx] = yac_pointer_to_unique_id(transient);
      ++transient_idx;
    }
  }
}

// determines whether a transient is source (can also be target)
int yac_is_transient_source(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  int src_component_id, tgt_component_id;
  yac_couple_config_get_transient_couple_component_ids(
    transient->couple_config, transient->couple_idx,
    transient->transient_couple_idx, &src_component_id, &tgt_component_id);

  return (transient->global_comp_id == src_component_id);
}

// determines whether a transient is target (can also be source)
int yac_is_transient_target(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  int src_component_id, tgt_component_id;
  yac_couple_config_get_transient_couple_component_ids(
    transient->couple_config, transient->couple_idx,
    transient->transient_couple_idx, &src_component_id, &tgt_component_id);

  return (transient->global_comp_id == tgt_component_id);
}

// gets the collection size of a transient
int yac_get_transient_collection_size(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  return
    (int)yac_couple_config_get_transient_collection_size(
      transient->couple_config, transient->couple_idx,
      transient->transient_couple_idx);
}

char const * yac_get_transient_coupling_period(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  return
    yac_couple_config_get_coupling_period(
      transient->couple_config, transient->couple_idx,
      transient->transient_couple_idx);
}

// gets the name of source grid used
char const * yac_get_transient_source_grid_name(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  char const * src_grid_name, * tgt_grid_name;
  yac_couple_config_get_transient_grid_names(
    transient->couple_config, transient->couple_idx,
    transient->transient_couple_idx, &src_grid_name, &tgt_grid_name);

  return src_grid_name;
}

// gets the name of target grid used
char const * yac_get_transient_target_grid_name(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  char const * src_grid_name, * tgt_grid_name;
  yac_couple_config_get_transient_grid_names(
    transient->couple_config, transient->couple_idx,
    transient->transient_couple_idx, &src_grid_name, &tgt_grid_name);

  return tgt_grid_name;
}

// gets the name of a transient
char const * yac_get_transient_name(int transient_id) {

  struct couple_config_transient * transient =
    (struct couple_config_transient *)
      yac_unique_id_to_pointer(transient_id, "transient_id");

  return
    yac_couple_config_get_transient_name(
      transient->couple_config, transient->couple_idx,
      transient->transient_couple_idx);
}
