!!
!! @file test_init_group.F90
!!
!! @copyright Copyright  (C)  2022 DKRZ, MPI-M
!!
!! @author Moritz Hanke <hanke@dkrz.de>
!!
!
! Keywords:
! Maintainer: Moritz Hanke <hanke@dkrz.de>
!             Rene Redler <rene.redler@mpimet.mpg.de>
! URL: https://dkrz-sw.gitlab-pages.dkrz.de/yac/
!
! This file is part of YAC.
!
! Redistribution and use in source and binary forms, with or without
! modification, are  permitted provided that the following conditions are
! met:
!
! Redistributions of source code must retain the above copyright notice,
! this list of conditions and the following disclaimer.
!
! Redistributions in binary form must reproduce the above copyright
! notice, this list of conditions and the following disclaimer in the
! documentation and/or other materials provided with the distribution.
!
! Neither the name of the DKRZ GmbH nor the names of its contributors
! may be used to endorse or promote products derived from this software
! without specific prior written permission.
!
! THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
! IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
! TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
! OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
! EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
! PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
! PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
! LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!


#include "test_macros.inc"

program test_init_group

  use utest
  use mo_yac_finterface
  use yaxt
  use mpi
  implicit none

  character(len=YAC_MAX_CHARLEN) :: xml_filename
  character(len=YAC_MAX_CHARLEN) :: xsd_filename

  integer :: global_rank, global_size
  integer :: ierror

  xml_filename = 'dummy_coupling.xml'
  xsd_filename = 'coupling.xsd'

  call mpi_init(ierror)

  call xt_initialize(mpi_comm_world)

  call mpi_comm_rank ( mpi_comm_world, global_rank, ierror )
  call mpi_comm_size ( mpi_comm_world, global_size, ierror )

  if (global_size /= 6) then
    write ( * , * ) "wrong number of processes (should be 6)"
    call error_exit
  endif

  call test_finit_group()
  call test_finit_comm_group()

  call xt_finalize()

  call mpi_finalize(ierror)

  call stop_test
  call exit_tests

contains

  ! testing of yac_finit_group and yac_finit_group_instance
  subroutine test_finit_group ()

    integer :: yac_instance_id
    logical :: is_dummy
    integer :: color

    integer :: ref_group_comm
    integer :: group_comm, group_comm_instance
    integer :: result, result_instance

    yac_instance_id = -1
    is_dummy = .false.
    color = mpi_undefined

    select case(global_rank)
      case (0:1)
        ! first two ranks are part of group A
        call yac_finit_group(xml_filename, xsd_filename, 'group a')
        call yac_finit_group_instance( &
          xml_filename, xsd_filename, 'group a', yac_instance_id)
        color = 0
      case (2:3)
        ! next two ranks are part of group B
        call yac_finit_group(xml_filename, xsd_filename, 'group b')
        call yac_finit_group_instance( &
          xml_filename, xsd_filename, 'group b', yac_instance_id)
        color = 1
      case (4)
        ! rank four does not provide a group
        call yac_finit_group(xml_filename, xsd_filename, '')
        call yac_finit_group_instance( &
          xml_filename, xsd_filename, '', yac_instance_id)
      case (5)
        ! rank five does a dummy initialisation
        call yac_finit_dummy()
        call yac_finit_dummy() ! instance
        is_dummy = .true.
    end select

    call mpi_comm_split(mpi_comm_world, color, 0, ref_group_comm, ierror)

    if (.not. is_dummy) then
      call yac_fget_groupcomm(group_comm)
      call yac_fget_groupcomm_instance(yac_instance_id, group_comm_instance)
    else
      group_comm = mpi_comm_null
      group_comm_instance = mpi_comm_null
    end if

    if (ref_group_comm /= mpi_comm_null) then

      call mpi_comm_compare(ref_group_comm, group_comm, result, ierror);
      call mpi_comm_compare( &
        ref_group_comm, group_comm_instance, result_instance, ierror);

      call test((result == MPI_CONGRUENT) .OR. (result == MPI_IDENT))
      call test((result_instance == MPI_CONGRUENT) .OR. (result_instance == MPI_IDENT))

      call MPI_Comm_free(ref_group_comm, ierror)

    else
      call test(group_comm == mpi_comm_null)
      call test(group_comm_instance == mpi_comm_null)
    end if

    if (.not. is_dummy) call yac_ffinalize_instance(yac_instance_id);
    call yac_ffinalize()

  end subroutine test_finit_group

  ! testing of yac_finit_comm_group and yac_finit_comm_group_instance
  subroutine test_finit_comm_group ()

    integer :: yac_instance_id
    integer :: world_comm, world_rank
    integer :: color

    integer :: ref_group_comm
    integer :: group_comm, group_comm_instance
    integer :: result, result_instance

    color = merge(mpi_undefined, 0, global_rank == 3)
    call mpi_comm_split(mpi_comm_world, color, 0, world_comm, ierror)

    if (global_rank == 3) return

    call mpi_comm_rank(world_comm, world_rank, ierror)

    select case(world_rank)
      case (0:1)
        ! first two ranks are part of group A
        call yac_finit_comm_group( &
          xml_filename, xsd_filename, world_comm, 'group a')
        call yac_finit_comm_group_instance( &
          xml_filename, xsd_filename, world_comm, 'group a', yac_instance_id)
        color = 0
      case (2:3)
        ! next two ranks are part of group B
        call yac_finit_comm_group( &
          xml_filename, xsd_filename, world_comm, 'group b')
        call yac_finit_comm_group_instance( &
          xml_filename, xsd_filename, world_comm, 'group b', yac_instance_id)
        color = 1
      case (4)
        ! rank four does not provide a group
        call yac_finit_comm(xml_filename, xsd_filename, world_comm)
        call yac_finit_comm_instance( &
          xml_filename, xsd_filename, world_comm, yac_instance_id)
        color = mpi_undefined
    end select

    call mpi_comm_split(world_comm, color, 0, ref_group_comm, ierror)

    call yac_fget_groupcomm(group_comm)
    call yac_fget_groupcomm_instance(yac_instance_id, group_comm_instance)

    if (ref_group_comm /= mpi_comm_null) then

      call mpi_comm_compare(ref_group_comm, group_comm, result, ierror);
      call mpi_comm_compare( &
        ref_group_comm, group_comm_instance, result_instance, ierror);

      call test((result == MPI_CONGRUENT) .OR. (result == MPI_IDENT))
      call test((result_instance == MPI_CONGRUENT) .OR. (result_instance == MPI_IDENT))

      call MPI_Comm_free(ref_group_comm, ierror)

    else
      call test(group_comm == mpi_comm_null)
      call test(group_comm_instance == mpi_comm_null)
    end if

    call yac_ffinalize_instance(yac_instance_id);
    call yac_ffinalize()

  end subroutine test_finit_comm_group

  subroutine error_exit ()

    use mpi, only : mpi_abort, mpi_comm_world
    use utest

    integer :: ierror

    call test ( .false. )
    call stop_test
    call exit_tests
    call mpi_abort ( mpi_comm_world, 999, ierror )

  end subroutine error_exit

end program test_init_group

